;============================================================
;
; DIS386 - An advanced disassembler for 80386/80486/Pentium CPUs
; Copyright (c) 1992-2003,2004 Alessandro Scotti
; http://www.ascotti.org/
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
; 25 Aug 1994 (A. Scotti):
;       BOUND and CMPXCHG8B are checked so that no size
;       information is added
;
;============================================================
.386
LOCALS
JUMPS

INCLUDE D3_DEFS.INC

PUBLIC  DIS3_Decode

;============================================================
;
; Data segment
;
DIS3_DATA       SEGMENT USE16 BYTE PUBLIC 'DATA'
                ASSUME ds:DIS3_DATA, es:NOTHING

INCLUDE D3_TABLE.INC

D3_Prefix       PFX_RECORD      ?       ; Prefix opcodes record

; Opcode informations
D3_Name         DB              ?       ; Instruction name
D3_Type         DB              ?       ; Instruction type (handler code)
D3_Arg1         DB              ?       ; First argument
D3_Arg2         DB              ?       ; Second argument
D3_Arg3         DB              ?       ; Third argument
D3_AuxOvrd      DB              ?       ; Auxiliary override segment

D3_Length       DW              ?       ; Instruction length in bytes
D3_ExtraLength  DW              ?       ; Extra bytes (e.g. prefixes)

D3_ValidTable	DW		?	; Offset of opcodes validation table

D3_Config       CNF_RECORD      ?       ; Configuration flags

D3_Opcode       DB      ?               ; Opcode
D3_OpcodeOfs    DW      ?               ; Opcode offset
D3_VirtualSeg   DW      ?               ; Opcode virtual offset (32 bit)
D3_VirtualOfs   DD      ?               ; Opcode virtual offset (32 bit)
D3_SavedVOfs    DD      ?               ; Opcode virtual offset (32 bit)

D3_OutBufPtr    LABEL   DWORD           ; Pointer to output buffer
D3_OutBufOfs    DW      ?
D3_OutBufSeg    DW      ?
D3_CurBufOfs    DW      ?               ; Current offset into buffer

D3_ListSize     DW      ?               ; Bytes reserved for listing

D3_MODRMreg     DW      ?               ; Offset of table for MODRM "reg" field
D3_MODRM        DB      ?               ; First byte of MODRM

; Table of opcode handlers
TBL_TypeHandler LABEL   WORD
        DW      SUB_0                   ; Error
        DW      SUB_1                   ; Standard
        DW      SUB_2                   ; Address size prefix
        DW      SUB_3                   ; Operand size prefix
        DW      SUB_4                   ; Jumps
        DW      SUB_5                   ; Segment override prefix
        DW      SUB_6                   ; Extended "imul"
        DW      SUB_7                   ; String prefix
        DW      SUB_8                   ; String operation
        DW      SUB_9                   ; Opcodes from MODRM "ttt" field
        DW      SUB_10                  ; Two-byte opcodes (prefix 0Fh)
        DW      SUB_11                  ; Coprocessor instructions
        DW      SUB_12                  ; "shld", "shrd"
        DW      SUB_13                  ; "aad", "aam"
        DW      SUB_14                  ; Word/dword instructions

; Table of addressing method handlers
TBL_AddrMethod  LABEL   WORD
        DW      SUB_A
        DW      SUB_C
        DW      SUB_D
        DW      SUB_E
        DW      SUB_G
        DW      SUB_I
        DW      SUB_J
        DW      SUB_M
        DW      SUB_O
        DW      SUB_R
        DW      SUB_S
        DW      SUB_T

DIS3_DATA       ENDS

;============================================================
;
; Code segment
;
DIS3_TEXT       SEGMENT USE16 BYTE PUBLIC 'CODE'
                ASSUME cs:DIS3_TEXT

;============================================================
;
; Disassembles a single instruction
;
; Input:
;       Control = address of control structure
; Output:
;       AX      = instruction length in bytes
;       DX      = 0 on success, 1 on error (invalid opcode)
; Notes:
;       control structure is updated.
;
DIS3_Decode     PROC    FAR
        ARG     Control:DWORD   = ARG_SIZE
        push    bp
        mov     bp, sp
        .push   ds, si, es, di, fs

        les     di, [Control]
        lds     si, es:[di].C_CodePtr
; Setup registers, save parameters
        push    ds
        push    ax
        mov     ax, SEG DIS3_DATA
        mov     ds, ax
        pop     ax

        mov     ax, es:[di].C_Flags
        mov     [D3_Config], ax
        mov     eax, es:[di].C_VirtualOfs
        mov     [D3_VirtualOfs], eax
        mov     [D3_SavedVOfs], eax
        mov     ax, es:[di].C_VirtualSeg
        mov     [D3_VirtualSeg], ax
        mov     ax, es:[di].C_ListSize
        mov     [D3_ListSize], ax

        mov     [D3_OpcodeOfs], si
        les     di, es:[di].C_BufferPtr
        mov     [D3_OutBufOfs], di
        mov     [D3_OutBufSeg], es
        pop     fs
        mov     bx, si                          ; FS:BX now points to opcode

; Setup variables
        call    SUB_ClearOutBuf
        mov     [D3_ExtraLength], 0
        mov     [D3_Prefix], 0
        test    [D3_Config], MASK CNF_USE32     ; 32-bit virtual offset?
        jnz     @@CheckCPUType                  ; Yes, continue
        and     [D3_VirtualOfs], 0000FFFFh      ; Clear high word
@@CheckCPUType:
        mov     ax, [D3_Config]
        and     ax, MASK CNF_CPU
        shr     ax, CNF_CPU
        mov     [D3_ValidTable], OFFSET TBL_AllOpcodes
        cmp     al, CNF_CPUALL
        je      @@Decode
        mov     [D3_ValidTable], OFFSET TBL_8086Opcodes
        cmp     al, CNF_CPU8086
        je      @@Decode
        mov     [D3_ValidTable], OFFSET TBL_80186Opcodes

; Decode loop
@@Decode:
        mov     [D3_Length], 1                  ; Assume 1-byte opcode
        mov     [D3_AuxOvrd], R_DS              ; Set aux override segment to DS
        mov     si, bx
        mov     al, fs:[si]                     ; Get opcode
        mov     [D3_Opcode], al
        movzx   si, al
        shl     si, 2                           ; Offset of opcode into table
        mov     ax, TBL_Opcodes[si]
        mov     [D3_Arg1], al                   ; Save first argument
        mov     [D3_Arg2], ah                   ; Save second argument
        mov     [D3_Arg3], ARG_NULL             ; Assume no other arguments
        mov     ax, TBL_Opcodes[si+02h]
        mov     [D3_Name], al                   ; Save name identifier
        mov     [D3_Type], ah                   ; Save handler identifier
        inc     bx                              ; Bump to next byte

; Check if opcode is valid
        mov     si, [D3_ValidTable]             ; Get offset of validation table
        movzx   ax, [D3_Opcode]
        bt      [si], ax                        ; Set CF if opcode is valid
        jnc     @@Error

; Call opcode handler
        movzx   si, [D3_Type]
        shl     si, 1
        call    TBL_TypeHandler[si]
        inc     [D3_VirtualOfs]                 ; Bump virtual offset
        xor     dx, dx                          ; Assume no errors
        or      al, al
        jz      @@Done
        cmp     al, RET_PREFIX
        je      @@Decode

; Error handler
@@Error:
        call    SUB_Error                       ; Handle error condition
        mov     dx, 1                           ; Return error

@@Done:
        .putc   0
        call    SUB_List                        ; Add listing
        les     di, [Control]
        movzx   eax, [D3_Length]
        add     ax, [D3_ExtraLength]
        add     es:[di].C_VirtualOfs, eax       ; Update virtual offset
        add     WORD PTR es:[di].C_CodePtr, ax  ; Update instruction offset

@@Exit:
        .pop    ds, si, es, di, fs
        mov     sp, bp
        pop     bp
        ret     ARG_SIZE
DIS3_Decode     ENDP

;============================================================
;
; Clears the output buffer
;
; Input:
;       none
; Output:
;       none
;
SUB_ClearOutBuf PROC    NEAR
        .push   cx, es, di
        mov     cx, BUFSIZE
        mov     es, [D3_OutBufSeg]
        mov     di, [D3_OutBufOfs]
        mov     [D3_CurBufOfs], di
        mov     al, ' '
        cld
        rep     stosb                           ; Clear output buffer
        mov     cx, [D3_ListSize]
        jcxz    @@Done
        mov     [D3_ListSize], 0
@@Loop:
        .putc   TAB
        add     [D3_ListSize], TABSIZE
        sub     cx, TABSIZE
        jg      @@Loop
@@Done:
        .pop    cx, es, di
        ret
SUB_ClearOutBuf ENDP

;============================================================
;
; Handles an error condition
;
; Input:
;       none
; Output:
;       none
;
SUB_Error       PROC    NEAR
        call    SUB_ClearOutBuf
; Write "DB"
        mov     si, OFFSET TBL_OpNames
        mov     al, N_DB
        call    SUB_PutSIdx
        .putc   TAB

; Write all bytes examined till now
        mov     cx, 1                           ; Skip just one opcode bytes...
        add     cx, [D3_ExtraLength]            ; ...but don't forget prefixes
        mov     bx, [D3_OpcodeOfs]              ; Initialize instruction pointer
        call    SUB_OutB                        ; Print the first byte
        dec     cx                              ; Any bytes left?
        jz      @@Done                          ; No, exit
@@Loop:
        .putc   ','                             ; Print a comma
        call    SUB_OutB
        loop    @@Loop                          ; Loop on remaining bytes

@@Done:
        mov     [D3_Length], 1                  ; Skip bad opcode
        ret
SUB_Error       ENDP

;============================================================
;
; Adds listing to the output buffer
;
; Input:
;       none
; Output:
;       none
;
SUB_List        PROC    NEAR
        push    [D3_Length]
        push    [D3_Config]
        and     [D3_Config], NOT (MASK CNF_SOURCE)

        mov     al, 4
        test    [D3_Config], MASK CNF_USE32
        jz      @@1
        add     al, 4
@@1:
        cbw
        add     ax, 7                           ; 'ssss:oooo(oooo)  '
        cmp     ax, [D3_ListSize]
        ja      @@Exit                          ; Not enough space, exit
        sub     [D3_ListSize], ax

; Write SEG:OFS
        mov     di, [D3_OutBufOfs]
        mov     [D3_CurBufOfs], di
        mov     ax, [D3_VirtualSeg]
        mov     cl, 4
        call    SUB_PutH
        .putc   ':'
        mov     eax, [D3_SavedVOfs]             ; Original virtual offset
        mov     cl, 4
        test    [D3_Config], MASK CNF_USE32
        jz      @@2
        mov     cl, 8
@@2:
        call    SUB_PutH
        .putc   ' '

; Compute number of bytes to write
        mov     ax, [D3_ListSize]
        sar     ax, 1                           ; Each byte is two digits
        jle     @@Exit                          ; Exit if out of space
        mov     cx, [D3_Length]
        add     cx, [D3_ExtraLength]            ; Number of bytes to write
        cmp     cx, ax                          ; Is there enough space?
        pushf                                   ; Save flags for later
        jbe     @@3                             ; Yes, continue
        mov     cx, ax
        dec     cx
@@3:
        mov     bx, [D3_OpcodeOfs]
@@Loop:
        call    SUB_OutB                        ; Write a byte
        loop    @@Loop
        popf                                    ; Restore flags
        jbe     @@Exit                          ; All chars done, exit
        .putc   '+'                             ; Not all bytes displayed

@@Exit:
        pop     [D3_Config]
        pop     [D3_Length]
        ret
SUB_List        ENDP

;============================================================
;
; Input:
;       AL      = index of string
;       DS:SI   = pointer to string table
; Output:
;       DS:SI   = pointer to selected string
;
SUB_ScanTable   PROC    NEAR
        .push   ax
        cld
        or      al, al                          ; First string?
        jz      @@Exit                          ; Yes, exit now
        mov     ah, al
@@Loop:
        lodsb
        or      al, al                          ; End of string?
        jns     @@Loop                          ; No, continue
        dec     ah                              ; Correct index?
        jnz     @@Loop                          ; No, search again
@@Exit:
        .pop    ax
        ret
SUB_ScanTable   ENDP

;============================================================
;
; Type 0: invalid opcode
;
SUB_0   PROC    NEAR
        mov     al, RET_ERROR
        ret
SUB_0   ENDP

;============================================================
;
; Type 1: standard
;
SUB_1   PROC    NEAR
        mov     al, fs:[bx]
        mov     [D3_MODRM], al

        call    SUB_1A
        ret
SUB_1   ENDP

;------------------------------------------------------------
;
; Scans and parses the argument list
;
SUB_1A  PROC    NEAR
        mov     al, [D3_Name]                   ; Get name code
        mov     si, OFFSET TBL_OpNames
        call    SUB_PutSIdx                     ; Write the name
        .putc   TAB                             ; Write a TAB

        movzx   ax, [D3_Arg1]
        call    SUB_ParseArg                    ; Parse arg1
        jc      @@Done                          ; Exit if error
        mov     al, [D3_Arg2]
        mov     ah, ','
        call    SUB_ParseArg                    ; Parse arg2
        jc      @@Done                          ; Exit if error
        mov     al, [D3_Arg3]
        mov     ah, ','
        call    SUB_ParseArg                    ; Parse arg3
        jc      @@Done                          ; Exit if error

        clc                                     ; Clear carry on success
@@Done:
        mov     al, 00h
        jnc     @@Exit
        mov     al, RET_ERROR
@@Exit:
        ret
SUB_1A  ENDP

;============================================================
;
; Type 2: address size prefix
;
SUB_2   PROC    NEAR
        or      [D3_Prefix], MASK PFX_ADDRESS32
        inc     [D3_ExtraLength]
        mov     al, RET_PREFIX
        ret
SUB_2   ENDP

;============================================================
;
; Type 3: operand size prefix
;
SUB_3   PROC    NEAR
        or      [D3_Prefix], MASK PFX_OPERAND32
        inc     [D3_ExtraLength]
        mov     al, RET_PREFIX
        ret
SUB_3   ENDP

;============================================================
;
; Type 4: standard jumps (only needs to check for indent on/off)
;
SUB_4   PROC    NEAR
        test    [D3_Config], MASK CNF_INDENTJMP ; Indent on?
        jz      @@Done                          ; No, exit
        .putc   ' '                             ; Write a space
@@Done:
        call    SUB_1
        ret
SUB_4   ENDP

;============================================================
;
; Type 5: segment override prefix
;
SUB_5   PROC    NEAR
        movzx   ax, [D3_Arg1]                   ; Get register name
        sub     al, R_ES                        ; Compute relative index
; Set segment override flag, leave handling to other procedures
        inc     [D3_ExtraLength]
        shl     ax, PFX_OVRDSEG                 ; Align field
        or      ax, MASK PFX_OVRD               ; Set segment override bit
        and     [D3_Prefix], NOT (MASK PFX_OVRDSEG)
        or      [D3_Prefix], ax                 ; Update prefix byte
        mov     al, RET_PREFIX
        ret
SUB_5   ENDP

;============================================================
;
; Type 6: IMUL Gv,Ev,Ib or Gv,Ev,Iv
;
SUB_6   PROC    NEAR
        mov     al, [D3_Arg1]                   ; Move arg1 to arg3
        mov     [D3_Arg3], al
        mov     [D3_Arg1], AM_G or OT_V         ; Set arg1
        mov     [D3_Arg2], AM_E or OT_V         ; Set arg2
        call    SUB_1                           ; Call standard parser
        ret
SUB_6   ENDP

;============================================================
;
; Type 7: repeat prefix
;
SUB_7   PROC    NEAR
        mov     ax, MASK PFX_REPE
        cmp     [D3_Name], N_REPE
        je      @@Done
        mov     ax, MASK PFX_REPNE
@@Done:
        and     [D3_Prefix], NOT (MASK PFX_REPE + MASK PFX_REPNE)
        or      [D3_Prefix], ax
        inc     [D3_ExtraLength]
        mov     al, RET_PREFIX
        ret
SUB_7   ENDP

;============================================================
;
; Type 8: string operation
;
SUB_8   PROC    NEAR
; Check for repeat prefix
        mov     ah, N_REPE
        test    [D3_Prefix], MASK PFX_REPE
        jnz     @@Rep
        mov     ah, N_REPNE
        test    [D3_Prefix], MASK PFX_REPNE
        jz      @@Main
; Write repeat prefix, always use "rep" for "lods", "stos" and "movs"
@@Rep:
        mov     al, [D3_Name]
        cmp     al, N_SCASB
        je      @@RepDone
        cmp     al, N_CMPSB
        je      @@RepDone
        mov     ah, N_REP               ; Use "rep" when flags don't matter
@@RepDone:
        xchg    al, ah
        mov     si, OFFSET TBL_OpNames
        call    SUB_PutSIdx
        .putc   TAB

; Write instruction name
@@Main:
        mov     al, [D3_Name]
        cmp     [D3_Arg1], OT_B
        je      @@MainDone
        inc     al
        test    [D3_Prefix], MASK PFX_OPERAND32
        jnz     @@MainDone
        inc     al
@@MainDone:
        mov     si, OFFSET TBL_OpNames
        call    SUB_PutSIdx

; Check for segment override prefix
        test    [D3_Prefix], MASK PFX_OVRD              ; Need override?
        jz      @@Exit                                  ; No, exit
        .putc   ' '
        mov     ax, [D3_Prefix]
        and     ax, MASK PFX_OVRDSEG
        shr     ax, PFX_OVRDSEG                         ; Get segment name
        mov     si, OFFSET TBL_RegsS
        call    SUB_PutSIdx                             ; Write segment name
        .putc   ':'

@@Exit:
        xor     al, al
        ret
SUB_8   ENDP

;============================================================
;
; Type 9: opcodes determined by bits 5,4,3 of MODRM field
;
SUB_9   PROC    NEAR
        LOCAL   bGroup:BYTE,                    \
                bOpIdx:BYTE     = VAR_SIZE
        enter   VAR_SIZE, 0

        mov     al, [D3_Name]
        mov     [bGroup], al
        mov     al, fs:[bx]
        and     al, 00111000b
        shr     al, 3
        mov     [bOpIdx], al
        mov     ah, [bGroup]                    ; Get group number
        dec     ah
        shl     ah, 3
        add     al, ah
        movzx   si, al
        mov     al, RET_ERROR
        cmp     BYTE PTR TBL_Groups[si], N_NULL ; Invalid opcode?
        je      @@Exit                          ; Yes, exit
        mov     al, TBL_Groups[si]
        mov     [D3_Name], al

; Adjust arguments if needed
        mov     al, [bGroup]                    ; Get group number
; Group 3
        cmp     al, 3                           ; Group 3?
        jne     @@NotGrp3                       ; No, continue
        mov     al, [D3_Arg1]                   ; Get 1st argument
        and     al, 01110000b                   ; Mask off addressing mode
        mov     ah, al
        or      al, AM_I                        ; Assume immediate data
        cmp     [bOpIdx], 000b
        je      @@Grp3Done                      ; Yes, done
        mov     al, ARG_NULL                    ; Assume no argument
@@Grp3Done:
        mov     [D3_Arg2], al
        jmp     @@Done
@@NotGrp3:
; Group 5
        cmp     al, 5
        jne     @@NotGrp5
        movzx   si, [bOpIdx]
        mov     al, TBL_Grp5Arg[si]
        mov     [D3_Arg1], al
        jmp     @@Done
@@NotGrp5:
; Group 6
        cmp     al, 6
        jne     @@NotGrp6
        mov     [D3_Arg1], AM_E or OT_W
        jmp     @@Done
@@NotGrp6:
; Group 7
        cmp     al, 7
        jne     @@NotGrp7
        movzx   si, [bOpIdx]
        mov     al, TBL_Grp7Arg[si]
        mov     [D3_Arg1], al
        jmp     @@Done
@@NotGrp7:
;
@@Done:
        call    SUB_1                           ; Parse name and arguments
@@Exit:
        leave
        ret
SUB_9   ENDP

;============================================================
;
; Type 10: two-byte opcodes (prefix 0Fh)
;
SUB_10  PROC    NEAR
        inc     [D3_ExtraLength]                ; Skip prefix byte
        mov     al, fs:[bx]                     ; Get 2nd byte
; Scan the list of two-byte opcodes
        mov     si, OFFSET TBL_Opcodes2
@@Loop:
        cmp     al, [si]
        je      @@Found
        add     si, 5
        cmp     si, OFFSET TBL_Opcodes2End
        jb      @@Loop
        mov     al, RET_ERROR
        jmp     @@Exit
@@Found:
; Initialize variables
        inc     si
        mov     ax, [si]
        mov     [D3_Arg1], al
        mov     [D3_Arg2], ah
        mov     al, [si+2]
        mov     [D3_Name], al
        movzx   si, BYTE PTR [si+3]
        shl     si, 1
        inc     bx                              ; Skip past opcode
        call    TBL_TypeHandler[si]
@@Exit:
        ret
SUB_10  ENDP

;============================================================
;
; Type 11: coprocessor
;
SUB_11  PROC    NEAR
        LOCAL   wOpWord:WORD,                   \
                bOpIdx:BYTE     = VAR_SIZE
        enter   VAR_SIZE, 0

        mov     al, fs:[bx]
        mov     ah, fs:[bx-1]
        and     ax, 0000011111111111b           ; Mask off escape bits
        mov     [wOpWord], ax                   ; Save opcode word

        and     ax, 10111100000b
        cmp     ax, 00111100000b                ; Group 4?
        je      @@Group4                        ; Yes, continue

        mov     ax, [wOpWord]
        and     ax, 11000000b
        cmp     ax, 11000000b                   ; Group 3?
        je      @@Group3                        ; Yes, continue

        mov     ax, [wOpWord]
        and     ax, 00100100000b
        cmp     ax, 00100100000b                ; Group 2?
        jne     @@Group2                        ; Yes, continue

;------------------------------------------------------------
; Group 1: 11011 OPA 1 MOD 1 OPB R/M
@@Group1:
        mov     ax, [wOpWord]
        and     ah, 00000110b                   ; First bits of opcode
        and     al, 00011000b                   ; Other bits of opcode
        shl     ah, 1
        shr     al, 3
        or      al, ah                          ; Opcode
        cbw
        shl     ax, 1
        add     ax, OFFSET TBL_Cop1
        mov     si, ax
        mov     ax, [si]
        push    ax
        call    SUB_11A
        pop     ax
        jc      @@Error
        mov     al, ah
        cmp     al, SI_NULL
        je      @@Group1Done
        mov     si, OFFSET TBL_SizeInfo
        call    SUB_PutSIdx
@@Group1Done:
        call    SUB_E                           ; Parse the MODRM byte
        jc      @@Error                         ; Exit if error
        xor     al, al
        jmp     @@Exit

;------------------------------------------------------------
; Group 2: 11011 MF OPA MOD OPB R/M
@@Group2:
        mov     ax, [wOpWord]
        and     ah, 00000001b                   ; First bit of opcode
        and     al, 00111000b                   ; Other bits of opcode
        shr     al, 3
        shl     ah, 3
        or      al, ah                          ; Opcode
        test    [wOpWord], 01000000000b         ; Integer instruction?
        jz      @@Group2Name                    ; No, continue
        add     al, 12                          ; Skip to "integer" names
@@Group2Name:
        push    bx
        mov     bx, OFFSET TBL_Cop2             ; Names for group 2
        xlat
        pop     bx
        call    SUB_11A
        jc      @@Error

        mov     al, SI_DWORD                    ; Assume size is 32 bit
        test    [wOpWord], 10000000000b         ; Correct size?
        jz      @@Group2Size                    ; Yes, print size info
        mov     al, SI_QWORD
        test    [wOpWord], 01000000000b         ; Integer instruction?
        jz      @@Group2Size                    ; No, continue
        mov     al, SI_WORD
@@Group2Size:
        mov     si, OFFSET TBL_SizeInfo
        call    SUB_PutSIdx

        call    SUB_E                           ; Parse the MODRM byte
        jc      @@Error                         ; Exit if error
        xor     al, al
        jmp     @@Exit

;------------------------------------------------------------
; Group 3: 11011 D P OPA 11 OPB ST(i)
@@Group3:
        inc     [D3_Length]
        mov     ax, [wOpWord]
        push    ax
        and     al, 111b
        mov     [bOpIdx], al
        pop     ax
        and     ah, 00000111b                   ; First bits of opcode
        and     al, 00111000b                   ; Other bits of opcode
        shr     al, 3
        shl     ah, 3
        or      al, ah                          ; Opcode
        and     al, 00001111b
        and     ah, 00110000b
        mov     si, OFFSET TBL_Cop300
        cmp     ah, 00010000b                   ; Subgroup 01b?
        jne     @@Group3Std                     ; No, proceed
; Subgroup 01b
        cmp     al, 0101b                       ; Valid pattern?
        jne     @@Error                         ; No, exit
        mov     ax, [wOpWord]
        and     al, 111b
        cmp     al, 001b                        ; Valid code?
        jne     @@Error                         ; No, exit
        mov     al, N_FUCOMPP
        call    SUB_11A
        jc      @@Error
        xor     al, al
        jmp     @@Exit
; Standard group 3 parser (00b, 10b, 11b)
@@Group3Std:
        mov     si, OFFSET TBL_Cop300
        cmp     ah, 00000000b
        je      @@Group3Name
        mov     si, OFFSET TBL_Cop310
        cmp     ah, 00100000b
        je      @@Group3Name
        mov     si, OFFSET TBL_Cop311
@@Group3Name:
        cbw
        shl     ax, 1
        add     si, ax
        mov     ax, [si]                        ; Get name and code
        cmp     al, N_NULL                      ; Valid instruction?
        je      @@Error                         ; No, exit
        push    ax
        call    SUB_11A                         ; Write instruction name
        pop     ax
        jc      @@Error
        test    ah, 11111000b                   ; Special subcode?
        jnz     @@Group3Arg                     ; No, parse arguments
        cmp     ah, [bOpIdx]                    ; Does subcode match?
        jne     @@Error                         ; No, exit
        xor     al, al
        jmp     @@Exit
@@Group3Arg:
        and     ah, 00011000b
        mov     dh, [bOpIdx]
        xor     dl, dl
        cmp     ah, CO_STSTI
        je      @@Group3Done
        xchg    dl, dh
        cmp     ah, CO_STIST
        je      @@Group3Done
        mov     dh, 8
@@Group3Done:
        mov     ax, dx
        call    SUB_11B
        xor     al, al
        jmp     @@Exit

;------------------------------------------------------------
; Group 4: 11011 0 S 1111 OP
@@Group4:
        mov     ax, [wOpWord]
        and     al, 00011111b
        test    ah, 00000010b
        jz      @@Group4Done
        or      al, 00100000b
@@Group4Done:
        push    bx
        mov     bx, OFFSET TBL_Cop4
        xlat
        pop     bx
        inc     bx                              ; Opcode length is two bytes
        inc     [D3_Length]
        call    SUB_11A
        jc      @@Error
        xor     al, al
        jmp     @@Exit

@@Error:
        mov     al, RET_ERROR
@@Exit:
        leave
        ret
SUB_11  ENDP

;------------------------------------------------------------
;
; Writes the name of a coprocessor instruction
;
SUB_11A PROC    NEAR
        push    ax
        .putc   'f'
        pop     ax
        cmp     al, N_NULL
        stc
        je      @@Exit
        mov     si, OFFSET TBL_CopNames
        call    SUB_PutSIdx
        .putc   TAB
        clc
@@Exit:
        ret
SUB_11A ENDP

;------------------------------------------------------------
;
; Writes the arguments of a coprocessor instruction
;
; Input:
;       AL      = index of 1st argument (0..7)
;       AH      = index of 2nd argument (0..7, otherwise no argument)
;
SUB_11B PROC    NEAR
        call    @@PutArg
        mov     al, ah
        cmp     al, 8
        jae     @@Exit
        push    ax
        .putc   ','
        test    [D3_Config], MASK CNF_SOURCE
        jz      @@Done
        .putc   ' '
@@Done:
        pop     ax
        call    @@PutArg
@@Exit:
        ret
@@PutArg:
        push    ax
        push    ax
        .putc   's'
        .putc   't'
        pop     ax
        or      al, al
        jz      @@PutArgExit
        add     al, '0'
        push    ax
        .putc   '('
        pop     ax
        .putc
        .putc   ')'
@@PutArgExit:
        pop     ax
        ret
SUB_11B ENDP

;============================================================
;
; Type 12: "shld", "shrd"
;
SUB_12  PROC    NEAR
        mov     al, [D3_Arg1]                   ; Move arg1 to arg3
        mov     [D3_Arg3], al
        mov     [D3_Arg1], AM_E or OT_V         ; Set arg1
        mov     [D3_Arg2], AM_G or OT_V         ; Set arg2
        call    SUB_1                           ; Call standard parser
        ret
SUB_12  ENDP

;============================================================
;
; Type 13: aaa, aam
;
SUB_13  PROC    NEAR
        cmp     BYTE PTR fs:[bx], 0Ah           ; Standard type?
        jne     @@Exit                          ; No, print argument too
        mov     [D3_Arg1], ARG_NULL             ; Delete argument
        inc     [D3_Length]                     ; Skip one extra byte
@@Exit:
        call    SUB_1
        ret
SUB_13  ENDP

;============================================================
;
; Type 14: word/dword instructions
;
SUB_14  PROC    NEAR
        mov     si, OFFSET TBL_16to32
@@Loop:
        mov     al, [si]
        cmp     al, N_NULL
        je      @@Exit
        cmp     al, [D3_Name]
        je      @@Found
        add     si, 4
        jmp     @@Loop
@@Found:
        mov     cl, [si+3]
        mov     dx, 1
        shl     dx, cl
        test    [D3_Prefix], dx                 ; 32-bit version?
        jz      @@Exit                          ; No, exit
        mov     al, [si+1]
@@Exit:
        mov     [D3_Name], al
        movzx   ax, BYTE PTR [si+2]
        mov     [D3_Type], al
        mov     si, ax
        shl     si, 1
        call    TBL_TypeHandler[si]
        ret
SUB_14  ENDP

;============================================================
;
; Addressing method: direct address
;
SUB_A   PROC    NEAR
        .push   cx, edx

        push    ax
        mov     cx, 2                           ; Assume a 16-bit offset
        mov     edx, fs:[bx]
        test    [D3_Prefix], MASK PFX_ADDRESS32 ; Guessed right?
        jz      @@1                             ; Yes, continue
        shl     cx, 1                           ; 32-bit offset
@@1:
        add     bx, cx                          ; Skip past offset
        add     [D3_Length], cx
        pop     ax

        push    cx
        cmp     al, OT_P                        ; Far address?
        jne     @@Short                         ; No, continue
        call    SUB_OutW                        ; Write the segment
        .putc   ':'
        jmp     @@Offset

; In source mode, prefix an intrasegment address with the CS: override
@@Short:
        test    [D3_Config], MASK CNF_SOURCE
        jz      @@Offset
        .putc   'c'
        .putc   's'
        .putc   ':'

@@Offset:
        pop     cx
        shl     cx, 1
        mov     eax, edx
        call    SUB_PutH                        ; Write offset

        .pop    bx, cx, dx
        clc
        ret
SUB_A   ENDP

;============================================================
;
; Addressing method: "reg" field of MODRM is a control register
;
SUB_C   PROC    NEAR
        mov     [D3_MODRMreg], OFFSET TBL_RegsC
        call    SUB_ParseMODRMreg
        ret
SUB_C   ENDP

;============================================================
;
; Addressing method: "reg" field of MODRM is a debug register
;
SUB_D   PROC    NEAR
        mov     [D3_MODRMreg], OFFSET TBL_RegsD
        call    SUB_ParseMODRMreg
        ret
SUB_D   ENDP

;============================================================
;
; Addressing method: MODRM "mod" + "r/m"
;
; Note:
;       FS:BX must point to the MODRM byte.
;
SUB_E   PROC    NEAR
        LOCAL   bOpType:BYTE,           \
                bRegMem:BYTE,           \
                bScale:BYTE,            \
                bIndex:BYTE,            \
                bBase:BYTE,             \
                bMod:BYTE       = VAR_SIZE
        enter   VAR_SIZE, 0
        .push   cx, edx

; Initialize local variables for one-byte MODRM
        mov     [bOpType], al
        mov     al, fs:[bx]
        inc     bx
        inc     [D3_Length]
        mov     ah, al
        and     al, 00000111b
        mov     [bRegMem], al
        mov     al, ah
        shr     al, 6
        mov     [bMod], al

        test    [D3_Prefix], MASK PFX_ADDRESS32 ; 32-bit addressing?
        jnz     @@80386                         ; Yes, 80386 MODRM type

; One-byte MODRM (8086 style)
@@8086:
        cmp     [bMod], 11b                     ; Register?
        je      @@Register                      ; Yes, "r/m" field is a register
        mov     dx, fs:[bx]                     ; Assume a 16-bit displacement
        mov     cx, 2                           ; Size is word
        cmp     [bMod], 10b
        je      @@8086_Mem
        movsx   dx, BYTE PTR fs:[bx]            ; Assume an 8-bit displacement
        mov     cl, 1                           ; Size is byte
        cmp     [bMod], 01b
        je      @@8086_Mem
        xor     dx, dx                          ; No displacement...
        xor     cl, cl                          ; ...so no extra bytes
        cmp     [bRegMem], 110b                 ; Special direct address case?
        jne     @@8086_Mem                      ; No, continue
        call    SUB_O
        clc
        jmp     @@Exit
@@8086_Mem:
        mov     [D3_AuxOvrd], ARG_NULL          ; Clear aux segment override
        add     bx, cx                          ; Skip displacement data
        add     [D3_Length], cx
        call    SUB_MemoryRef
        mov     si, OFFSET TBL_AddrMode86
        mov     al, [bRegMem]
        call    SUB_PutSIdx
        or      dx, dx
        jz      @@8086_MemDone
        mov     al, '+'
        or      dx, dx
        jg      @@8086_Mem2
        mov     al, '-'
        neg     dx
@@8086_Mem2:
        .putc
        mov     ax, dx
        shl     cl, 1                           ; One byte, two digits
        call    SUB_PutH
@@8086_MemDone:
        .putc   ']'
        clc
        jmp     @@Exit
@@Register:
        mov     al, [bOpType]
        call    SUB_AdjustOperand
        mov     si, OFFSET TBL_Regs8
        cmp     al, OT_B
        je      @@Reg_Done
        mov     si, OFFSET TBL_Regs16
        cmp     al, OT_W
        je      @@Reg_Done
        mov     si, OFFSET TBL_Regs32
@@Reg_Done:
        mov     al, [bRegMem]
        call    SUB_PutSIdx
        clc
        jmp     @@Exit

; One or two-byte MODRM (80386 style)
@@80386:
        cmp     [bMod], 11b                     ; Register?
        je      @@Register                      ; Yes, "r/m" field is a register
        mov     [bScale], 00b                   ; Assume no scaling factor
        mov     [bIndex], 100b                  ; Assume no index register
        mov     al, [bRegMem]
        mov     [bBase], al
        cmp     al, 100b                        ; Escape to two-byte?
        jne     @@80386_Continue                ; No, continue
; Get MODRM second byte
        mov     al, fs:[bx]
        inc     bx
        inc     [D3_Length]
        mov     ah, al
        and     al, 00000111b
        mov     [bBase], al
        mov     al, ah
        shr     al, 3
        and     al, 00111b
        mov     [bIndex], al
        mov     al, ah
        shr     al, 6
        mov     [bScale], al
@@80386_Continue:
        mov     edx, fs:[bx]                    ; Assume a 32-bit displacement
        mov     cx, 4                           ; Size is double word
        cmp     [bMod], 10b
        je      @@80386_Mem
        mov     dl, fs:[bx]                     ; Assume an 8-bit displacement
        mov     cl, 1                           ; Size is byte
        cmp     [bMod], 01b
        je      @@80386_Mem
        xor     dx, dx                          ; No displacement...
        xor     cl, cl                          ; ...so no extra bytes
        cmp     [bBase], 101b                   ; Special direct address case?
        jne     @@80386_Mem                     ; No, continue
        call    SUB_O
        clc
        jmp     @@Exit
@@80386_Mem:
        mov     [D3_AuxOvrd], ARG_NULL          ; Clear aux segment override
        add     bx, cx                          ; Skip displacement data
        add     [D3_Length], cx
        call    SUB_MemoryRef
        mov     si, OFFSET TBL_Regs32
        mov     al, [bBase]
        call    SUB_PutSIdx
        cmp     [bIndex], 100b                  ; Index register?
        je      @@80386_IdxDone                 ; No, skip this part
        .putc   '+'
        mov     si, OFFSET TBL_Regs32
        mov     al, [bIndex]
        call    SUB_PutSIdx
        cmp     [bScale], 00b                   ; Scale factor?
        je      @@80386_IdxDone                 ; No, skip this part
        .putc   '*'
        push    cx
        mov     al, 1
        mov     cl, [bScale]
        shl     al, cl
        pop     cx
        add     al, '0'
        .putc
        mov     [bScale], 00b                   ; Clear scale factor
@@80386_IdxDone:
        cmp     [bScale], 00b                   ; At this point, [bScale]
        jne     @@Error                         ; should be zero
        or      dx, dx                          ; Any displacement?
        jz      @@80386_MemDone                 ; No, done
        .putc   '+'
        mov     eax, edx                        ; Get displacement
        shl     cl, 1                           ; One byte, two digits
        call    SUB_PutH
@@80386_MemDone:
        .putc   ']'
        clc
        jmp     @@Exit

@@Error:
        stc
@@Exit:
        .pop    cx, edx
        leave
        ret
SUB_E   ENDP

;============================================================
;
; Addressing method: "reg" field of MODRM is a general register
;
SUB_G   PROC    NEAR
        call    SUB_AdjustOperand
        mov     [D3_MODRMreg], OFFSET TBL_Regs8
        cmp     al, OT_B
        je      @@Done
        mov     [D3_MODRMreg], OFFSET TBL_Regs16
        cmp     al, OT_W
        je      @@Done
        mov     [D3_MODRMreg], OFFSET TBL_Regs32
@@Done:
        call    SUB_ParseMODRMreg
        ret
SUB_G   ENDP

;============================================================
;
; Addressing method: immediate data
;
SUB_I   PROC    NEAR
        .push   dx
        call    SUB_AdjustOperand               ; Get effective operand size

        mov     dx, OFFSET SUB_OutB
        cmp     al, OT_B
        je      @@Done
        mov     dx, OFFSET SUB_OutW
        cmp     al, OT_W
        je      @@Done
        mov     dx, OFFSET SUB_OutD
        cmp     al, OT_D
        je      @@Done
        stc                                     ; Return an error
        jmp     @@Exit
@@Done:
        call    dx
        clc
@@Exit:
        .pop    dx
        ret
SUB_I   ENDP

;============================================================
;
; Addressing method: relative offset
;
SUB_J   PROC    NEAR
        .push   ecx, edx
        call    SUB_AdjustOperand

        movsx   edx, BYTE PTR fs:[bx]
        mov     cx, 1
        cmp     al, OT_B                        ; Short jump?
        je      @@Continue                      ; Ok, proceed
        movsx   edx, WORD PTR fs:[bx]
        mov     cl, 2
        cmp     al, OT_W                        ; 16-bit offset?
        je      @@Continue                      ; Ok, proceed
        mov     edx, fs:[bx]
        mov     cl, 4                           ; 32-bit offset
@@Continue:
        add     bx, cx                          ; Skip past data
        add     [D3_Length], cx                 ; Update length field

        test    [D3_Config], MASK CNF_SOURCE    ; Source mode?
        jnz     @@Source                        ; Yes, proceed

; Compute destination offset
        movzx   eax, [D3_Length]
        add     eax, [D3_VirtualOfs]
        add     eax, edx
        mov     cl, 4                           ; Assume a 32-bit offset
        test    [D3_Prefix], MASK PFX_OPERAND32 ; Is it really a 32-bit offset?
        jnz     @@Print                         ; Yes, print it
        mov     cl, 2                           ; 16-bit offset
        jmp     @@Print

; Source mode: use the "$+displacement" form (note: size in CX is preserved)
@@Source:
; Patch zone: LOOPD, LOOPDE, LOOPDNE and JECXZ are assembled with an address
; size prefix, so that length is three bytes
        cmp     [D3_Name], N_JECXZ
        je      @@Patch
        cmp     [D3_Name], N_LOOPD
        je      @@Patch
        cmp     [D3_Name], N_LOOPDE
        je      @@Patch
        cmp     [D3_Name], N_LOOPDNE
        jne     @@SrcDone
@@Patch:
        inc     [D3_Length]                     ; Bump instruction length
        dec     [D3_ExtraLength]                ; Decrement extra length
@@SrcDone:
        movzx   eax, [D3_Length]                ; Get instruction length...
        add     edx, eax                        ; and add it to displacement
        .putc   '$'
        mov     al, '+'                         ; Assume positive displacement
        or      edx, edx
        jge     @@SrcPrint                      ; Ok if not negative
        mov     al, '-'                         ; Invert sign
        neg     edx                             ; Get absolute value
@@SrcPrint:
        .putc                                   ; Print sign
        mov     eax, edx                        ; Copy displacement into EAX

@@Print:
        shl     cl, 1                           ; One byte, two digits!
        call    SUB_PutH

@@Exit:
        clc
        .pop    ecx, edx
        ret
SUB_J   ENDP

;============================================================
;
; Addressing method: the MODRM field may refer only to memory
;
SUB_M   PROC    NEAR
        mov     ah, fs:[bx]
        and     ah, 11000000b                   ; Mask "mod" field
        cmp     ah, 11000000b                   ; General register?
        stc                                     ; Assume an error
        je      @@Exit                          ; Exit if error
        call    SUB_E                           ; Parse MODRM as usual
@@Exit:
        ret
SUB_M   ENDP

;============================================================
;
; Addressing method: direct offset
;
SUB_O   PROC    NEAR
        .push   dx

        mov     dx, OFFSET SUB_OutW             ; Assume a 16-bit address
        test    [D3_Prefix], MASK PFX_ADDRESS32 ; Address size prefix?
        jz      @@Done                          ; No, so offset is 16-bit
        mov     dx, OFFSET SUB_OutD

@@Done:
	call	SUB_MemoryRef
        call    dx
        .putc   ']'

        clc
        .pop    dx
        ret
SUB_O   ENDP

;============================================================
;
; Addressing method: "mod" field of MODRM must refer to a general register
;
SUB_R   PROC    NEAR
        mov     ah, fs:[bx]
        and     ah, 11000000b                   ; Mask "mod" field
        cmp     ah, 11000000b                   ; General register?
        stc                                     ; Assume an error
        jne     @@Exit                          ; Exit if error
        call    SUB_E                           ; Parse MODRM as usual
@@Exit:
        ret
SUB_R   ENDP

;============================================================
;
; Addressing method: "reg" field of MODRM is a segment register
;
SUB_S   PROC    NEAR
        mov     [D3_MODRMreg], OFFSET TBL_RegsS
        call    SUB_ParseMODRMreg
        ret
SUB_S   ENDP

;============================================================
;
; Addressing method: "reg" field of MODRM is a test register
;
SUB_T   PROC    NEAR
        mov     [D3_MODRMreg], OFFSET TBL_RegsT
        call    SUB_ParseMODRMreg
        ret
SUB_T   ENDP

;============================================================
;
; Parses the MODRM "reg" field
;
; Input:
;       none
; Output:
;       CF      = set on error
;
SUB_ParseMODRMreg       PROC    NEAR
        mov     si, [D3_MODRMreg]
        mov     al, [D3_MODRM]
        shr     al, 3
        and     al, 07h
        call    SUB_ScanTable
        cmp     BYTE PTR [si], N_NULL           ; Is selected entry valid?
        je      @@Exit                          ; No, exit (note that CF=0)
        call    SUB_PutS                        ; Write the string
        stc                                     ; Set carry
@@Exit:
        cmc                                     ; Complement carry
        ret
SUB_ParseMODRMreg       ENDP

;============================================================
;
; Outputs the "memory reference" header: size, segment override and
; square bracket
;
; Input:
;       none
; Output:
;	none
;
SUB_MemoryRef	PROC	NEAR
        .push   ax, dx, si, bp

; Check whether size information is needed
        cmp     [D3_Type], 11                   ; Coprocessor instruction?
        je      @@Override                      ; Yes, skip size info
        mov     al, [D3_Name]
        cmp     al, N_BOUND                     ; BOUND?
        je      @@Override                      ; Yes, no info
        cmp     al, N_CMPXCHG8B                 ; CMPXCHG8B?
        je      @@Override                      ; Yes, no info
        xor     dx, dx                          ; Information counter
        mov     al, [D3_Arg1]
        test    al, 80h
        jnz     @@Arg2                          ; No info needed for arg1
        and     al, 00001111b
        cmp     al, AM_E                        ; MODRM?
        je      @@Arg1Info                      ; Yes, need info
        cmp     al, AM_I                        ; Immediate data?
        je      @@Arg1Info                      ; Yes, need info
        cmp     al, AM_M                        ; Memory reference?
        jne     @@Arg2                          ; No, check arg2
@@Arg1Info:
        inc     dl
        inc     dh                              ; Arg1 need help!
@@Arg2:
        mov     al, [D3_Arg2]
        cmp     al, ARG_NULL                    ; Null argument?
        je      @@Arg2Info                      ; Yes, may need info!
        test    al, 80h
        jnz     @@Opcode                        ; No info needed for arg2
        and     al, 00001111b
        cmp     al, AM_E                        ; MODRM?
        je      @@Arg2Info                      ; Yes, need info
        cmp     al, AM_I                        ; Immediate data?
        je      @@Arg2Info                      ; Yes, need info
        cmp     al, AM_M                        ; Memory reference?
        jne     @@Opcode                        ; No, check opcode
@@Arg2Info:
        inc     dl
@@Opcode:
        inc     dl                              ; Assume info needed
        mov     al, [D3_Name]
        cmp     al, N_MOVSX                     ; Sign extend?
        je      @@SizeInfo                      ; Yes, need info
        cmp     al, N_MOVZX                     ; Zero extend?
        je      @@SizeInfo                      ; Yes, need info
        dec     dl                              ; Decrement info counter
@@SizeInfo:
        cmp     dl, 2                           ; So, really need info?
        jb      @@Override                      ; No, skip this part
        mov     al, [D3_Arg1]                   ; Check arg1
        or      dh, dh                          ; Useful info here?
        jnz     @@SizeInfo1                     ; Yes, use this argument
        mov     al, [D3_Arg2]                   ; Check arg2
        test    al, 80h                         ; Useful info here?
        jnz     @@Override                      ; No, skip this section
@@SizeInfo1:
        and     al, 01110000b                   ; Mask operand type
        call    SUB_AdjustOperand
        mov     ah, SI_BYTE
        cmp     al, OT_B
        je      @@SizeDone
        mov     ah, SI_WORD
        cmp     al, OT_W
        je      @@SizeDone
        mov     ah, SI_DWORD
        cmp     al, OT_D
        je      @@SizeDone
        mov     ah, SI_QWORD
        cmp     al, OT_Q
        je      @@SizeDone
        mov     ah, SI_FWORD
        cmp     al, OT_S
        je      @@SizeDone
        test    [D3_Prefix], MASK PFX_OPERAND32 ; Must be OT_P... 48 bit?
        jnz     @@SizeDone                      ; Yes, done
        mov     ah, SI_DWORD                    ; It's 32 bit
@@SizeDone:
        mov     si, OFFSET TBL_SizeInfo
        mov     al, ah
        call    SUB_PutSIdx

; Check whether a segment override is needed
@@Override:
        mov     ax, [D3_Prefix]
        and     ax, MASK PFX_OVRDSEG
        shr     ax, PFX_OVRDSEG
        test    [D3_Prefix], MASK PFX_OVRD
        jnz     @@OvrdDone
        test    [D3_Config], MASK CNF_SOURCE    ; Source mode?
        jz      @@Done                          ; No, don't write the override
        mov     al, [D3_AuxOvrd]                ; Use the auxiliary override
        cmp     al, ARG_NULL
        je      @@Done
        sub     al, R_ES
@@OvrdDone:
        mov     si, OFFSET TBL_RegsS
        call    SUB_PutSIdx
        .putc   ':'

@@Done:
	.putc	'['
        .pop    ax, dx, si, bp
	ret
SUB_MemoryRef	ENDP

;============================================================
;
; Adjusts an operand type which may depend on the operand size prefix
;
; Input:
;       AL      = operand type (B, C, D, V, W)
; Output:
;       AL      = operand type (B, W, D)
;
SUB_AdjustOperand       PROC    NEAR
        and     al, 01110000b                   ; Mask operand bits
        cmp     al, OT_V                        ; Word/dword type?
        je      @@V
        cmp     al, OT_C                        ; Byte/word type?
        je      @@C
        jmp     @@Exit
@@V:
        mov     al, OT_W                        ; Assume word size
        test    [D3_Prefix], MASK PFX_OPERAND32 ; Operand size prefix?
        jz      @@Exit                          ; No, exit
        mov     al, OT_D                        ; Size is double word
        jmp     @@Exit
@@C:
        mov     al, OT_B                        ; Assume byte size
        test    [D3_Prefix], MASK PFX_OPERAND32 ; Operand size prefix?
        jz      @@Exit                          ; No, exit
        mov     al, OT_W                        ; Size is word
        jmp     @@Exit
@@Exit:
        ret
SUB_AdjustOperand       ENDP

;============================================================
;
; Parses an argument byte
;
; Input:
;       AH      = ',' if argument must be preceeded by a comma
;       AL      = argument byte
; Output:
;       CF      = set on error
;
SUB_ParseArg    PROC    NEAR
        cmp     al, ARG_NULL                    ; Null argument?
        je      @@Done                          ; Yes, exit (note that CF=0)
        cmp     ah, ','                         ; Must write a comma?
        jne     @@Continue                      ; No, continue
        xchg    al, ah
        call    SUB_PutC
        xchg    al, ah
        test    [D3_Config], MASK CNF_SOURCE    ; Source mode?
        jz      @@Continue                      ; No, continue
        push    ax
        .putc   ' '                             ; Write a blank
        pop     ax
@@Continue:
        test    al, 80h                         ; Alternate argument set?
        jz      @@Primary                       ; No, continue

; Alternate set: register and special constants
        cmp     al, R_1                         ; Special value?
        jne     @@Not1                          ; No, continue
        .putc   '1'
        clc
        jmp     @@Done
@@Not1:
        mov     si, OFFSET TBL_Regs8
        cmp     al, R_ES
        jae     @@RegDone
        cmp     al, R_EAX
        jb      @@RegDone
        test    [D3_Prefix], MASK PFX_OPERAND32
        jnz     @@RegDone
        sub     al, 08h
@@RegDone:
        and     al, 00011111b
        call    SUB_PutSIdx
        clc
        jmp     @@Done

; Primary set: bits 3 thru 0 contain the addressing method, and
; bits 6 thru 4 the operand type
@@Primary:
        mov     si, ax
        and     si, 000Fh
        shl     si, 1
        and     al, 01110000b
        call    TBL_AddrMethod[si]

@@Done:
        ret
SUB_ParseArg    ENDP

;============================================================
;
; Writes a character into the output buffer (handles TAB)
;
; Input:
;       AL      = character
; Output:
;       none
;
SUB_PutC        PROC    NEAR
        .push   es, di

        mov     di, [D3_CurBufOfs]
        mov     es, [D3_OutBufSeg]

        test    [D3_Config], MASK CNF_UPCASE    ; Convert to uppercase?
        jz      @@Done                          ; No, continue
        cmp     al, 'a'
        jb      @@Done
        cmp     al, 'z'
        ja      @@Done
        sub     al, 'a'-'A'

@@Done:
        cmp     al, TAB
        jne     @@ASCII

@@TabLoop:
        inc     [D3_CurBufOfs]
        mov     di, [D3_CurBufOfs]
        sub     di, [D3_OutBufOfs]
        test    di, TABSIZE-1
        jz      @@Exit
        jmp     @@TabLoop

@@ASCII:
        mov     es:[di], al                     ; Put character into the buffer
        inc     [D3_CurBufOfs]                  ; Bump buffer offset

@@Exit:
        .pop    es, di
        ret
SUB_PutC        ENDP

;============================================================
;
; Writes a string into the output buffer
;
; Input:
;       DS:SI   = pointer to source string
; Output:
;       none
;
SUB_PutS        PROC    NEAR
        cld
@@Loop:
        lodsb                                   ; Get next character
        or      al, al                          ; End of string?
        js      @@Done                          ; Yes, break loop
        call    SUB_PutC
        jmp     @@Loop
@@Done:
        and     al, 7Fh                         ; Reset high byte
        jz      @@Exit                          ; Exit if null character
        call    SUB_PutC
@@Exit:
        ret
SUB_PutS        ENDP

;============================================================
;
; Writes an indexed string into the output buffer
;
; Input:
;       AL      = index of string
;       DS:SI   = pointer to string table
; Output:
;       none
;
SUB_PutSIdx     PROC    NEAR
        call    SUB_ScanTable
        call    SUB_PutS
        ret
SUB_PutSIdx     ENDP

;============================================================
;
; Writes an exadecimal number into the output buffer
;
; Input:
;       EAX     = value to convert
;       CL      = number of hex digits to output
; Output:
;       none
;
SUB_PutH        PROC    NEAR
        .push   cx, edx
        or      cl, cl
        jz      @@Exit
        mov     ch, cl

        mov     edx, eax
@@Loop1:
        mov     eax, edx
        shr     edx, 4
        and     al, 0Fh
; Fast nibble (0?h) to hex-digit conversion:
; if( AL <= 9 ) then
;       AL      = 9?h and AF=0  after "add al, 90h"
;       AL      = 9?h and CF=0  after "daa" (no changes)
;       AL      = D?h and CF=0  after "adc al, 40h" (note: AF is still zero)
;       AL      = 3?h           after "daa" (30h is the ASCII character '0')
; else ( AL > 9, e.g. AL = 0Bh)
;       AL      = 9Bh           after "add al, 90h" (AF doesn't matter here)
;       AL      = 01h and CF=1  after "daa" (add 66h to AL)
;       AL      = 42h           after "adc al, 40h" (add 40h, then carry)
;       AL      = 42h           after "daa" (42h is the ASCII character 'B')
; Note: another interesting method is the sequence
;       daa
;       add     al, F0h
;       adc     al, 40h
; but it needs AF = 0 on entry to work.
        add     al, 90h
        daa
        adc     al, 40h
        daa
        push    ax
        dec     cl
        jnz     @@Loop1

; In source mode, an hex number which starts with a letter
; must be preceeded by a leading zero
        test    [D3_Config], MASK CNF_SOURCE    ; Source mode?
        jz      @@Loop2                         ; No, jump to the second part
        pop     ax                              ; Pop the first digit
        push    ax                              ; Push it back onto the stack
        cmp     al, 'A'                         ; Is it a letter?
        jb      @@Loop2                         ; No, jump to the second part
        .putc   '0'                             ; Print a leading zero

@@Loop2:
        pop     ax
        call    SUB_PutC
        dec     ch
        jnz     @@Loop2

; In source mode, an hex number is followed by 'h'
        test    [D3_Config], MASK CNF_SOURCE    ; Source mode?
        jz      @@Exit                          ; No, exit
        push    [D3_Config]
        and     [D3_Config], NOT MASK CNF_UPCASE
        .putc   'h'                             ; Print hex number identifier
        pop     [D3_Config]

@@Exit:
        .pop    cx, edx
        ret
SUB_PutH        ENDP

;============================================================
;
; Prints the byte at FS:BX
;
; Input:
;       none
; Output:
;       FS:BX   = pointer to data past current byte
;
SUB_OutB        PROC    NEAR
        .push   ax, cx
        mov     al, fs:[bx]
        inc     bx
        inc     [D3_Length]
        mov     cl, 2
        call    SUB_PutH
        .pop    ax, cx
        ret
SUB_OutB        ENDP

;============================================================
;
; Prints the word at FS:BX
;
; Input:
;       none
; Output:
;       FS:BX   = pointer to data past current word
;
SUB_OutW        PROC    NEAR
        .push   ax, cx
        mov     ax, fs:[bx]
        inc     bx
        inc     bx
        add     [D3_Length], 2
        mov     cl, 4
        call    SUB_PutH
        .pop    ax, cx
        ret
SUB_OutW        ENDP

;============================================================
;
; Prints the double word at FS:BX
;
; Input:
;       none
; Output:
;       FS:BX   = pointer to data past current dword
;
SUB_OutD        PROC    NEAR
        .push   ax, cx
        mov     eax, fs:[bx]
        add     bx, 4
        add     [D3_Length], 4
        mov     cl, 8
        call    SUB_PutH
        .pop    ax, cx
        ret
SUB_OutD        ENDP

DIS3_TEXT       ENDS
END
